@@ -1,108 +1,92 @@
-# -*- coding:utf-8 -*-
 import json
 from abc import ABC, abstractmethod
-from io import BytesIO
 from pathlib import Path
-from struct import unpack
-from typing import List, Optional, Union
-
+from typing import Dict, Iterable, List, Optional, Tuple, Union
 import numpy as np
 from h3.api import numpy_int as h3
 
-from timezonefinder import utils
+from timezonefinder.np_binary_helpers import (
+    get_zone_ids_path,
+    get_zone_positions_path,
+    read_per_polygon_vector,
+)
+from timezonefinder.polygon_array import PolygonArray
+from timezonefinder import utils, utils_clang
 from timezonefinder.configs import (
-    BINARY_DATA_ATTRIBUTES,
-    BINARY_FILE_ENDING,
-    DATA_ATTRIBUTE_NAMES,
-    DTYPE_FORMAT_H,
-    DTYPE_FORMAT_H_NUMPY,
-    DTYPE_FORMAT_I,
-    DTYPE_FORMAT_SIGNED_I_NUMPY,
-    HOLE_ADR2DATA,
-    HOLE_COORD_AMOUNT,
-    HOLE_DATA,
-    HOLE_REGISTRY,
-    HOLE_REGISTRY_FILE,
-    NR_BYTES_H,
-    NR_BYTES_I,
-    POLY_ADR2DATA,
-    POLY_COORD_AMOUNT,
-    POLY_DATA,
-    POLY_MAX_VALUES,
-    POLY_NR2ZONE_ID,
-    POLY_ZONE_IDS,
-    SHORTCUT_FILE,
+    DEFAULT_DATA_DIR,
     SHORTCUT_H3_RES,
-    TIMEZONE_NAMES_FILE,
     CoordLists,
     CoordPairs,
 )
-from timezonefinder.hex_helpers import read_shortcuts_binary
-from timezonefinder.utils import inside_polygon
+
+from timezonefinder.flatbuf.shortcut_utils import (
+    get_shortcut_file_path,
+    read_shortcuts_binary,
+)
+from timezonefinder.zone_names import read_zone_names
 
 
 class AbstractTimezoneFinder(ABC):
-    # TODO document attributes in all classes
     # prevent dynamic attribute assignment (-> safe memory)
+    """
+    Abstract base class for TimezoneFinder instances
+    """
+
     __slots__ = [
-        "bin_file_location",
+        "data_location",
         "shortcut_mapping",
         "in_memory",
         "_fromfile",
         "timezone_names",
-        POLY_ZONE_IDS,
+        "zone_ids",
+        "holes_dir",
+        "boundaries_dir",
+        "boundaries",
+        "holes",
     ]
-    binary_data_attributes: List[str] = [POLY_ZONE_IDS]
+
+    zone_ids: np.ndarray
+    """
+    List of attribute names that store opened binary data files.
+    """
 
     def __init__(
         self,
         bin_file_location: Optional[Union[str, Path]] = None,
         in_memory: bool = False,
     ):
-        self.in_memory = in_memory
-
-        if self.in_memory:
-            self._fromfile = utils.fromfile_memory
-        else:
-            self._fromfile = np.fromfile
-
-        # open all the files in binary reading mode
-        # for more info on what is stored in which .bin file, please read the comments in file_converter.py
+        """
+        Initialize the AbstractTimezoneFinder.
+        :param bin_file_location: The path to the binary data files to use. If None, uses native package data.
+        :param in_memory: ignored. The binary files will be read into memory (few MB)
+        """
         if bin_file_location is None:
-            bin_file_location = Path(__file__).parent
-        else:
-            bin_file_location = Path(bin_file_location)
-        self.bin_file_location: Path = bin_file_location
+            bin_file_location = DEFAULT_DATA_DIR
+        self.data_location: Path = Path(bin_file_location)
 
-        with open(self.bin_file_location / TIMEZONE_NAMES_FILE, "r") as json_file:
-            self.timezone_names = json.loads(json_file.read())
+        self.timezone_names = read_zone_names(self.data_location)
 
-        path2shortcut_bin = self.bin_file_location / SHORTCUT_FILE
+        path2shortcut_bin = get_shortcut_file_path(self.data_location)
         self.shortcut_mapping = read_shortcuts_binary(path2shortcut_bin)
 
-        for attribute_name in self.binary_data_attributes:
-            file_name = attribute_name + BINARY_FILE_ENDING
-            path2file = self.bin_file_location / file_name
-            if self.in_memory:
-                with open(path2file, mode="rb") as bin_file:
-                    bf_in_mem = BytesIO(bin_file.read())
-                    bf_in_mem.seek(0)
-                setattr(self, attribute_name, bf_in_mem)
-            else:
-                bin_file = open(path2file, mode="rb")
-                setattr(self, attribute_name, bin_file)
-
-    def __del__(self):
-        for attribute_name in self.binary_data_attributes:
-            getattr(self, attribute_name).close()
+        zone_ids_path = get_zone_ids_path(self.data_location)
+        self.zone_ids = read_per_polygon_vector(zone_ids_path)
 
     @property
     def nr_of_zones(self):
+        """
+        Get the number of timezones.
+
+        :rtype: int
+        """
         return len(self.timezone_names)
 
     @staticmethod
     def using_numba() -> bool:
         """
+        Check if Numba is being used.
+
+        :rtype: bool
         :return: True if Numba is being used to JIT compile helper functions
         """
         return utils.using_numba
@@ -112,39 +96,73 @@ class AbstractTimezoneFinder(ABC):
         """
         :return: True if the compiled C implementation of the point in polygon algorithm is being used
         """
-        return utils.inside_polygon == utils.pt_in_poly_clang
+        return utils.inside_polygon == utils_clang.pt_in_poly_clang
 
     def zone_id_of(self, poly_id: int) -> int:
-        poly_zone_ids = getattr(self, POLY_ZONE_IDS)
-        poly_zone_ids.seek(NR_BYTES_H * poly_id)
-        return unpack(DTYPE_FORMAT_H, poly_zone_ids.read(NR_BYTES_H))[0]
+        """
+        Get the zone ID of a polygon.
+
+        :param poly_id: The ID of the polygon.
+        :type poly_id: int
+        :rtype: int
+        """
+        try:
+            return self.zone_ids[poly_id]
+        except TypeError:
+            raise ValueError(f"zone_ids is not set in directory {self.data_location}.")
 
     def zone_ids_of(self, poly_ids: np.ndarray) -> np.ndarray:
-        poly_zone_ids = getattr(self, POLY_ZONE_IDS)
-        id_array = np.empty(shape=len(poly_ids), dtype=DTYPE_FORMAT_H_NUMPY)
+        """
+        Get the zone IDs of multiple polygons.
 
-        for i, poly_id in enumerate(poly_ids):
-            poly_zone_ids.seek(NR_BYTES_H * poly_id)
-            id_array[i] = unpack(DTYPE_FORMAT_H, poly_zone_ids.read(NR_BYTES_H))[0]
-
-        return id_array
+        :param poly_ids: An array of polygon IDs.
+        :return: array of zone IDs corresponding to the polygon IDs.
+        """
+        return self.zone_ids[poly_ids]
 
     def zone_name_from_id(self, zone_id: int) -> str:
+        """
+        Get the zone name from a zone ID.
+
+        :param zone_id: The ID of the zone.
+        :return: The name of the zone.
+        :raises ValueError: If the timezone could not be found.
+        """
         try:
             return self.timezone_names[zone_id]
         except IndexError:
             raise ValueError("timezone could not be found. index error.")
 
     def zone_name_from_poly_id(self, poly_id: int) -> str:
+        """
+        Get the zone name from a polygon ID.
+
+        :param poly_id: The ID of the polygon.
+        :return: The name of the zone.
+        """
         zone_id = self.zone_id_of(poly_id)
         return self.zone_name_from_id(zone_id)
 
     def get_shortcut_polys(self, *, lng: float, lat: float) -> np.ndarray:
-        hex_id = h3.geo_to_h3(lat, lng, SHORTCUT_H3_RES)
+        """
+        Get the polygon IDs in the shortcut corresponding to the given coordinates.
+
+        :param lng: The longitude of the point in degrees (-180.0 to 180.0).
+        :param lat: The latitude of the point in degrees (90.0 to -90.0).
+        :return: An array of polygon IDs.
+        """
+        hex_id = h3.latlng_to_cell(lat, lng, SHORTCUT_H3_RES)
         shortcut_poly_ids = self.shortcut_mapping[hex_id]
         return shortcut_poly_ids
 
     def most_common_zone_id(self, *, lng: float, lat: float) -> Optional[int]:
+        """
+        Get the most common zone ID in the shortcut corresponding to the given coordinates.
+
+        :param lng: The longitude of the point in degrees (-180.0 to 180.0).
+        :param lat: The latitude of the point in degrees (90.0 to -90.0).
+        :return: The most common zone ID or None if no polygons exist in the shortcut.
+        """
         polys = self.get_shortcut_polys(lng=lng, lat=lat)
         if len(polys) == 0:
             return None
@@ -155,7 +173,11 @@ class AbstractTimezoneFinder(ABC):
 
     def unique_zone_id(self, *, lng: float, lat: float) -> Optional[int]:
         """
-        :return: the zone id at the coordinate if there is exactly one possible zone, else `None`
+        Get the unique zone ID in the shortcut corresponding to the given coordinates.
+
+        :param lng: The longitude of the point in degrees (-180.0 to 180.0).
+        :param lat: The latitude of the point in degrees (90.0 to -90.0).
+        :return: The unique zone ID or None if no polygons exist in the shortcut.
         """
         polys = self.get_shortcut_polys(lng=lng, lat=lat)
         if len(polys) == 0:
@@ -247,90 +269,123 @@ class TimezoneFinder(AbstractTimezoneFinder):
     :ivar binary_data_attributes: the names of all attributes which store the opened binary data files
 
     :param bin_file_location: path to the binary data files to use, None if native package data should be used
-    :param in_memory: whether to completely read and keep the binary files in memory
+    :param in_memory: whether to completely read and keep the timezone polygon binary files in memory
     """
 
     # __slots__ declared in parents are available in child classes. However, child subclasses will get a __dict__
     # and __weakref__ unless they also define __slots__ (which should only contain names of any additional slots).
-    __slots__ = DATA_ATTRIBUTE_NAMES
+    __slots__ = [
+        "hole_registry",
+        "_boundaries_file",
+        "_holes_file",
+    ]
 
-    binary_data_attributes = BINARY_DATA_ATTRIBUTES
-
-    def __init__(self, bin_file_location: Optional[str] = None, in_memory: bool = False):
-        super(TimezoneFinder, self).__init__(bin_file_location, in_memory)
+    def __init__(
+        self, bin_file_location: Optional[str] = None, in_memory: bool = False
+    ):
+        super().__init__(bin_file_location, in_memory)
+        self.holes_dir = utils.get_holes_dir(self.data_location)
+        self.boundaries_dir = utils.get_boundaries_dir(self.data_location)
+        self.boundaries = PolygonArray(
+            data_location=self.boundaries_dir, in_memory=in_memory
+        )
+        self.holes = PolygonArray(data_location=self.holes_dir, in_memory=in_memory)
 
         # stores for which polygons (how many) holes exits and the id of the first of those holes
-        # since there are very few it is feasible to keep them in the memory
-        with open(self.bin_file_location / HOLE_REGISTRY_FILE, "r") as json_file:
-            hole_registry_tmp = json.loads(json_file.read())
+        # since there are very few entries it is feasible to keep them in the memory
+        self.hole_registry = self._load_hole_registry()
 
+    def _load_hole_registry(self) -> Dict[int, Tuple[int, int]]:
+        """
+        Load and convert the hole registry from JSON file, converting keys to int.
+        """
+        path = utils.get_hole_registry_path(self.data_location)
+        with open(path, encoding="utf-8") as json_file:
+            hole_registry_tmp = json.loads(json_file.read())
         # convert the json string keys to int
-        hole_registry = {int(k): v for k, v in hole_registry_tmp.items()}
-        setattr(self, HOLE_REGISTRY, hole_registry)
+        return {int(k): v for k, v in hole_registry_tmp.items()}
 
     @property
-    def nr_of_polygons(self) -> int:
-        poly_zone_ids = getattr(self, POLY_ZONE_IDS)
-        return utils.get_file_size_byte(poly_zone_ids) // NR_BYTES_H
+    def nr_of_polygons(self):
+        return len(self.boundaries)
+
+    @property
+    def nr_of_holes(self):
+        return len(self.holes)
 
     def coords_of(self, polygon_nr: int = 0) -> np.ndarray:
-        poly_coord_amount = getattr(self, POLY_COORD_AMOUNT)
-        poly_adr2data = getattr(self, POLY_ADR2DATA)
-        poly_data = getattr(self, POLY_DATA)
+        """
+        Get the coordinates of a polygon from the FlatBuffers collection.
 
-        # how many coordinates are stored in this polygon
-        poly_coord_amount.seek(NR_BYTES_I * polygon_nr)
-        nr_of_values = unpack(DTYPE_FORMAT_I, poly_coord_amount.read(NR_BYTES_I))[0]
+        :param polygon_nr: The index of the polygon.
+        :return: Array of coordinates.
+        """
+        return self.boundaries.coords_of(polygon_nr)
 
-        poly_adr2data.seek(NR_BYTES_I * polygon_nr)
-        poly_data.seek(unpack(DTYPE_FORMAT_I, poly_adr2data.read(NR_BYTES_I))[0])
-        return np.stack(
-            (
-                self._fromfile(poly_data, dtype=DTYPE_FORMAT_SIGNED_I_NUMPY, count=nr_of_values),
-                self._fromfile(poly_data, dtype=DTYPE_FORMAT_SIGNED_I_NUMPY, count=nr_of_values),
-            )
-        )
-
-    def _holes_of_poly(self, polygon_nr: int):
-        hole_coord_amount = getattr(self, HOLE_COORD_AMOUNT)
-        hole_adr2data = getattr(self, HOLE_ADR2DATA)
-        hole_data = getattr(self, HOLE_DATA)
-        hole_registry = getattr(self, HOLE_REGISTRY)
+    def _iter_hole_ids_of(self, polygon_nr: int) -> Iterable[int]:
+        """
+        Yield the hole IDs for a given polygon number.
 
+        :param polygon_nr: Number of the polygon
+        :yield: Hole IDs
+        """
         try:
-            amount_of_holes, first_hole_id = hole_registry[polygon_nr]
+            amount_of_holes, first_hole_id = self.hole_registry[polygon_nr]
         except KeyError:
             return
+        for i in range(amount_of_holes):
+            yield first_hole_id + i
 
-        hole_coord_amount.seek(NR_BYTES_H * first_hole_id)
-        hole_adr2data.seek(NR_BYTES_I * first_hole_id)
+    def _holes_of_poly(self, polygon_nr: int):
+        """
+        Get the holes of a polygon from the FlatBuffers collection.
 
-        for _ in range(amount_of_holes):
-            nr_of_values = unpack(DTYPE_FORMAT_H, hole_coord_amount.read(NR_BYTES_H))[0]
-            hole_data.seek(unpack(DTYPE_FORMAT_I, hole_adr2data.read(NR_BYTES_I))[0])
+        :param polygon_nr: Number of the polygon
+        :yield: Generator of hole coordinates
+        """
+        for hole_id in self._iter_hole_ids_of(polygon_nr):
+            yield self.holes.coords_of(hole_id)
 
-            x_coords = self._fromfile(hole_data, dtype=DTYPE_FORMAT_SIGNED_I_NUMPY, count=nr_of_values)
-            y_coords = self._fromfile(hole_data, dtype=DTYPE_FORMAT_SIGNED_I_NUMPY, count=nr_of_values)
-            yield np.array(
-                [
-                    x_coords,
-                    y_coords,
-                ]
-            )
+    def get_polygon(
+        self, polygon_nr: int, coords_as_pairs: bool = False
+    ) -> List[Union[CoordPairs, CoordLists]]:
+        """
+        Get the polygon coordinates of a given polygon number.
 
-    def get_polygon(self, polygon_nr: int, coords_as_pairs: bool = False) -> List[Union[CoordPairs, CoordLists]]:
+        :param polygon_nr: Polygon number
+        :param coords_as_pairs: Determines the structure of the polygon representation
+        :return: List of polygon coordinates
+        """
         list_of_converted_polygons = []
         if coords_as_pairs:
             conversion_method = utils.convert2coord_pairs
         else:
             conversion_method = utils.convert2coords
-        list_of_converted_polygons.append(conversion_method(self.coords_of(polygon_nr=polygon_nr)))
+        list_of_converted_polygons.append(
+            conversion_method(self.coords_of(polygon_nr=polygon_nr))
+        )
 
         for hole in self._holes_of_poly(polygon_nr):
             list_of_converted_polygons.append(conversion_method(hole))
 
         return list_of_converted_polygons
 
+    def _iter_poly_ids_of_zone(self, zone_id: int) -> Iterable[int]:
+        """
+        Yield the polygon IDs for a given zone ID.
+
+        :param zone_id: ID of the zone
+        :yield: Polygon IDs
+        """
+        # load only on demand. used only in get_geometry() which is a non performance critical utility function
+        zone_positions_path = get_zone_positions_path(self.data_location)
+        zone_positions = np.load(zone_positions_path, mmap_mode="r")
+        first_poly_id_zone = zone_positions[zone_id]
+        # read the poly_id of the first polygon of the consequent zone
+        # NOTE: this has also been added for the last zone
+        first_poly_id_next = zone_positions[zone_id + 1]
+        yield from range(first_poly_id_zone, first_poly_id_next)
+
     def get_geometry(
         self,
         tz_name: Optional[str] = "",
@@ -354,67 +409,70 @@ class TimezoneFinder(AbstractTimezoneFinder):
             if not isinstance(tz_id, int):
                 raise TypeError("the zone id must be given as int.")
             if tz_id < 0 or tz_id >= self.nr_of_zones:
-                raise ValueError(f"the given zone id {tz_id} is invalid (value range: 0 - {self.nr_of_zones - 1}.")
+                raise ValueError(
+                    f"the given zone id {tz_id} is invalid (value range: 0 - {self.nr_of_zones - 1}."
+                )
         else:
+            if tz_name is None:
+                raise ValueError("no timezone name given.")
             try:
                 tz_id = self.timezone_names.index(tz_name)
             except ValueError:
                 raise ValueError("The timezone '", tz_name, "' does not exist.")
         if tz_id is None:
             raise ValueError("no timezone id given.")
-        poly_id2zone_id = getattr(self, POLY_NR2ZONE_ID)
-        poly_id2zone_id.seek(NR_BYTES_H * tz_id)
-        # read poly_id of the first polygon of that zone
-        this_zone_poly_id = unpack(DTYPE_FORMAT_H, poly_id2zone_id.read(NR_BYTES_H))[0]
-        # read poly_id of the first polygon of the consequent zone
-        # (also exists for the last zone, cf. file_converter.py)
-        next_zone_poly_id = unpack(DTYPE_FORMAT_H, poly_id2zone_id.read(NR_BYTES_H))[0]
-        # read and return all polygons from this zone:
-        return [self.get_polygon(poly_id, coords_as_pairs) for poly_id in range(this_zone_poly_id, next_zone_poly_id)]
 
-    def outside_the_boundaries_of(self, poly_id: int, x: int, y: int) -> bool:
-        # get the boundaries of the polygon = (lng_max, lng_min, lat_max, lat_min) converted to int32
-        poly_max_values = getattr(self, POLY_MAX_VALUES)
-        poly_max_values.seek(4 * NR_BYTES_I * poly_id)
-        xmax, xmin, ymax, ymin = self._fromfile(
-            poly_max_values,
-            dtype=DTYPE_FORMAT_SIGNED_I_NUMPY,
-            count=4,
-        )
-        return x > xmax or x < xmin or y > ymax or y < ymin
+        return [
+            self.get_polygon(poly_id, coords_as_pairs)
+            for poly_id in self._iter_poly_ids_of_zone(tz_id)
+        ]
+
+    def get_polygon_boundaries(self, poly_id: int) -> Tuple[int, int, int, int]:
+        """returns the bounding box of the polygon = (lng_max, lng_min, lat_max, lat_min) converted to int32"""
+        xmax = self.boundaries.xmax[poly_id]
+        xmin = self.boundaries.xmin[poly_id]
+        ymax = self.boundaries.ymax[poly_id]
+        ymin = self.boundaries.ymin[poly_id]
+        return xmax, xmin, ymax, ymin
 
     def inside_of_polygon(self, poly_id: int, x: int, y: int) -> bool:
-        # only read polygon (hole) data on demand
-        # only run the expensive algorithm if the point is withing the boundaries
-        if self.outside_the_boundaries_of(poly_id, x, y):
+        """
+        Check if a point is inside a polygon.
+
+        :param poly_id: Polygon ID # TODO rename to boundary_id
+        :param x: X-coordinate of the point
+        :param y: Y-coordinate of the point
+        :return: True if the point is inside the polygon, False otherwise
+        """
+        # avoid running the expensive PIP algorithm at any cost
+        # -> check bboxes first
+        if self.boundaries.outside_bbox(poly_id, x, y):
             return False
 
-        if not inside_polygon(x, y, self.coords_of(polygon_nr=poly_id)):
+        # NOTE: holes are much smaller -> less expensive to check
+        # -> check holes before the polygon
+        hole_id_iter = self._iter_hole_ids_of(poly_id)
+        if self.holes.in_any_polygon(hole_id_iter, x, y):
+            # the point is within a hole of the polygon
+            # it is excluded fromn the this boundary polygon
             return False
 
-        # when the point is within a hole of the polygon, this timezone must not be returned
-        if any(iter(inside_polygon(x, y, hole) for hole in self._holes_of_poly(poly_id))):
-            return False
-
-        # the query point is included in this polygon, but not any hole
-        return True
+        return self.boundaries.pip(poly_id, x, y)
 
     def timezone_at(self, *, lng: float, lat: float) -> Optional[str]:
-        """computes in which ocean OR land timezone a point is included in
+        """
+        Find the timezone for a given point, considering both land and ocean timezones.
 
-        Especially for large polygons it is expensive to check if a point is really included.
-        In case there is only one possible zone (left), this zone will instantly be returned without actually checking
-        if the query point is included in this polygon.
+        Uses precomputed shortcuts to reduce the number of polygons checked. Returns the timezone name
+        of the matched polygon, which may be an ocean timezone ("Etc/GMT+-XX") if applicable.
 
-        To speed things up there are "shortcuts" being used
-            which have been precomputed and store which timezone polygons have to be checked.
+        Since ocean timezones span the whole globe, some timezone will always be matched!
+        `None` can only be returned when using custom timezone data without such ocean timezones.
 
-        .. note:: Since ocean timezones span the whole globe, some timezone will always be matched!
-            `None` can only be returned when you have compiled timezone data without such "full coverage".
 
-        :param lng: longitude of the point in degree (-180.0 to 180.0)
-        :param lat: latitude in degree (90.0 to -90.0)
-        :return: the timezone name of the matched timezone polygon. possibly "Etc/GMT+-XX" in case of an ocean timezone.
+        :param lng: longitude of the point in degrees (-180.0 to 180.0)
+        :param lat: latitude of the point in degrees (90.0 to -90.0)
+        :return: the timezone name of the matched polygon, or None if no match is found.
         """
         lng, lat = utils.validate_coordinates(lng, lat)
         possible_polygons = self.get_shortcut_polys(lng=lng, lat=lat)
@@ -442,7 +500,6 @@ class TimezoneFinder(AbstractTimezoneFinder):
 
         # check until the point is included in one of the possible polygons
         for i, poly_id in enumerate(possible_polygons):
-
             if i >= last_zone_change_idx:
                 break
 
@@ -452,6 +509,7 @@ class TimezoneFinder(AbstractTimezoneFinder):
 
         # since it is the last possible option,
         # the polygons of the last possible zone don't actually have to be checked
+        # -> instantly return the last zone
         zone_id = zone_ids[-1]
         return self.zone_name_from_id(zone_id)
 
@@ -460,13 +518,13 @@ class TimezoneFinder(AbstractTimezoneFinder):
 
         .. note:: this is only meaningful when you have compiled your own timezone data
             where there are areas without timezone polygon coverage.
-            Otherwise some timezone will always be matched and the functionality is equal to using `.timezone_at()`
+            Otherwise, some timezone will always be matched and the functionality is equal to using `.timezone_at()`
             -> useless to actually test all polygons.
 
         .. note:: using this function is less performant than `.timezone_at()`
 
         :param lng: longitude of the point in degree
-        :param lat: latitude in degree
+        :param lat: latitude of the point in degree
         :return: the timezone name of the polygon the point is included in or `None`
         """
         lng, lat = utils.validate_coordinates(lng, lat)
